@@ -1,9 +1,10 @@
 cr = "\n"
 
 s = "####################################################################" + cr
-s += "# FINANCEPY BETA Version " + str('0.22') + " - This build:  21 Dec 2021 at 15:56 #" + cr
+s += "# FINANCEPY BETA Version " + str('0.182') + " - This build:  01 Oct 2020 at 18:23 #" + cr
 s += "#      This software is distributed FREE & WITHOUT ANY WARRANTY    #" + cr
-s += "#  Report bugs as issues at https://github.com/domokane/FinancePy  #" + cr
+s += "# For info and disclaimer - https://github.com/domokane/FinancePy  #" + cr
+s += "#      Send any bug reports or comments to quant@financepy.com     #" + cr
 s += "####################################################################"
 s += cr
 
