@@ -1,56 +1,53 @@
-# -*- coding: utf-8 -*-
-"""
-Created on Fri Feb 12 16:51:05 2016
+##############################################################################
+# Copyright (C) 2018, 2019, 2020 Dominic O'Kane
+##############################################################################
 
-@author: Dominic O'Kane
-"""
 
 import numpy as np
 from numba import njit, float64, int64
 from math import exp, log
+from copy import deepcopy
 
 from ...finutils.FinDate import FinDate
+from ...finutils.FinError import FinError
 from ...finutils.FinCalendar import FinCalendar, FinCalendarTypes
-from ...finutils.FinCalendar import FinDayAdjustTypes, FinDateGenRuleTypes
+from ...finutils.FinCalendar import FinBusDayAdjustTypes, FinDateGenRuleTypes
 from ...finutils.FinDayCount import FinDayCount, FinDayCountTypes
 from ...finutils.FinFrequency import FinFrequency, FinFrequencyTypes
 from ...finutils.FinGlobalVariables import gDaysInYear
 from ...finutils.FinMath import ONE_MILLION
-from ...market.curves.FinInterpolate import FinInterpMethods, uinterpolate
-from ...market.curves.FinDiscountCurve import FinDiscountCurve
+from ...finutils.FinHelperFunctions import labelToString, tableToString
+from ...market.curves.FinInterpolate import FinInterpTypes, _uinterpolate
+
+from ...finutils.FinHelperFunctions import checkArgumentTypes
 
 useFlatHazardRateIntegral = True
 standardRecovery = 0.40
 
-##########################################################################
+###############################################################################
+# TODO: Perform protection leg pv analytically using fact that hazard rate and
+#       interest rates are flat between their combined node points. Right now I
+#       do not find the protection leg PV calculations to be a bottleneck,
+#       especially given the speedup benefits of using NUMBA.
+###############################################################################
 
 
-@njit(
-    float64[:](
-        float64,
-        float64,
-        float64[:],
-        float64[:],
-        float64[:],
-        float64[:],
-        float64[:],
-        float64[:],
-        int64),
-    fastmath=True,
-    cache=True)
-def riskyPV01_NUMBA(teff,
-                    accrualFactorPCDToNow,
-                    paymentTimes,
-                    yearFracs,
-                    npLiborTimes,
-                    npLiborValues,
-                    npSurvTimes,
-                    npSurvValues,
-                    pv01Method):
+@njit(float64[:](float64, float64, float64[:], float64[:], float64[:],
+                 float64[:], float64[:], float64[:], int64),
+      fastmath=True, cache=True)
+def _riskyPV01_NUMBA(teff,
+                     accrualFactorPCDToNow,
+                     paymentTimes,
+                     yearFracs,
+                     npLiborTimes,
+                     npLiborValues,
+                     npSurvTimes,
+                     npSurvValues,
+                     pv01Method):
     ''' Fast calculation of the risky PV01 of a CDS using NUMBA.
     The output is a numpy array of the full and clean risky PV01.'''
 
-    method = FinInterpMethods.FLAT_FORWARDS.value
+    method = FinInterpTypes.FLAT_FORWARDS.value
 
     couponAccruedIndicator = 1
 
@@ -62,11 +59,11 @@ def riskyPV01_NUMBA(teff,
 
     # The first coupon is a special case which needs to be handled carefully
     # taking into account what coupon has already accrued and what has not
-    qeff = uinterpolate(teff, npSurvTimes, npSurvValues, method)
-    q1 = uinterpolate(tncd, npSurvTimes, npSurvValues, method)
-    z1 = uinterpolate(tncd, npLiborTimes, npLiborValues, method)
+    qeff = _uinterpolate(teff, npSurvTimes, npSurvValues, method)
+    q1 = _uinterpolate(tncd, npSurvTimes, npSurvValues, method)
+    z1 = _uinterpolate(tncd, npLiborTimes, npLiborValues, method)
 
-    # this is the part of the coupon accrued from the previous coupon date to now
+    # this is the part of the coupon accrued from previous coupon date to now
     # accrualFactorPCDToNow = dayCount.yearFrac(pcd,teff)
 
     # reference credit survives to the premium payment date
@@ -80,14 +77,15 @@ def riskyPV01_NUMBA(teff,
     # future accrued from now to coupon payment date assuming default roughly
     # midway
     fullRPV01 += 0.5 * z1 * \
-        (qeff - q1) * (yearFracs[1] - accrualFactorPCDToNow) * couponAccruedIndicator
+        (qeff - q1) * (yearFracs[1] - accrualFactorPCDToNow) \
+        * couponAccruedIndicator
 
     for it in range(2, len(paymentTimes)):
 
         t2 = paymentTimes[it]
 
-        q2 = uinterpolate(t2, npSurvTimes, npSurvValues, method)
-        z2 = uinterpolate(t2, npLiborTimes, npLiborValues, method)
+        q2 = _uinterpolate(t2, npSurvTimes, npSurvValues, method)
+        z2 = _uinterpolate(t2, npLiborTimes, npLiborValues, method)
 
         accrualFactor = yearFracs[it]
 
@@ -95,7 +93,7 @@ def riskyPV01_NUMBA(teff,
         # payment date
         fullRPV01 += q2 * z2 * accrualFactor
 
-        ####################################################################
+        #######################################################################
 
         if couponAccruedIndicator == 1:
 
@@ -114,7 +112,7 @@ def riskyPV01_NUMBA(teff,
 
             fullRPV01 = fullRPV01 + dfullRPV01
 
-        ####################################################################
+        #######################################################################
 
         q1 = q2
 
@@ -122,38 +120,38 @@ def riskyPV01_NUMBA(teff,
 
     return np.array([fullRPV01, cleanRPV01])
 
-##########################################################################
+###############################################################################
 
 @njit(float64(float64, float64, float64[:], float64[:], float64[:], float64[:],
               float64, int64, int64), fastmath=True, cache=True)
-def protectionLegPV_NUMBA(teff,
-                          tmat,
-                          npLiborTimes,
-                          npLiborValues,
-                          npSurvTimes,
-                          npSurvValues,
-                          contractRecovery,
-                          numStepsPerYear,
-                          protMethod):
+def _protectionLegPV_NUMBA(teff,
+                           tmat,
+                           npLiborTimes,
+                           npLiborValues,
+                           npSurvTimes,
+                           npSurvValues,
+                           contractRecovery,
+                           numStepsPerYear,
+                           protMethod):
     ''' Fast calculation of the CDS protection leg PV using NUMBA to speed up
     the numerical integration over time. '''
 
-    method = FinInterpMethods.FLAT_FORWARDS.value
+    method = FinInterpTypes.FLAT_FORWARDS.value
     dt = (tmat - teff) / numStepsPerYear
     t = teff
-    z1 = uinterpolate(t, npLiborTimes, npLiborValues, method)
-    q1 = uinterpolate(t, npSurvTimes, npSurvValues, method)
+    z1 = _uinterpolate(t, npLiborTimes, npLiborValues, method)
+    q1 = _uinterpolate(t, npSurvTimes, npSurvValues, method)
 
     protPV = 0.0
     small = 1e-8
 
     if useFlatHazardRateIntegral is True:
 
-        for i in range(0, numStepsPerYear):
+        for _ in range(0, numStepsPerYear):
 
             t = t + dt
-            z2 = uinterpolate(t, npLiborTimes, npLiborValues, method)
-            q2 = uinterpolate(t, npSurvTimes, npSurvValues, method)
+            z2 = _uinterpolate(t, npLiborTimes, npLiborValues, method)
+            q2 = _uinterpolate(t, npSurvTimes, npSurvValues, method)
             # This needs to be updated to handle small h+r
             h12 = -log(q2 / q1) / dt
             r12 = -log(z2 / z1) / dt
@@ -166,11 +164,11 @@ def protectionLegPV_NUMBA(teff,
 
     else:
 
-        for i in range(0, numStepsPerYear):
+        for _ in range(0, numStepsPerYear):
 
             t += dt
-            z2 = uinterpolate(t, npLiborTimes, npLiborValues, method)
-            q2 = uinterpolate(t, npSurvTimes, npSurvValues, method)
+            z2 = _uinterpolate(t, npLiborTimes, npLiborValues, method)
+            q2 = _uinterpolate(t, npSurvTimes, npSurvValues, method)
             dq = q1 - q2
             dprotPV = 0.5 * (z1 + z2) * dq
             protPV += dprotPV
@@ -180,8 +178,9 @@ def protectionLegPV_NUMBA(teff,
     protPV = protPV * (1.0 - contractRecovery)
     return protPV
 
-##########################################################################
-##########################################################################
+###############################################################################
+###############################################################################
+###############################################################################
 
 
 class FinCDS(object):
@@ -189,61 +188,35 @@ class FinCDS(object):
     generation and the valuation and risk management of CDS. '''
 
     def __init__(self,
-                 stepInDate, #  FinDate is when protection starts (usually T+1)
-                 maturityDateOrTenor,  # FinDate or a FinTenor
-                 runningCoupon, # Annualised coupon on premium leg
-                 notional=ONE_MILLION,
-                 longProtection=True,
-                 frequencyType=FinFrequencyTypes.QUARTERLY,
-                 dayCountType=FinDayCountTypes.ACT_360,
-                 calendarType=FinCalendarTypes.WEEKEND,
-                 busDayAdjustType=FinDayAdjustTypes.FOLLOWING,
-                 dateGenRuleType=FinDateGenRuleTypes.BACKWARD):
+                 stepInDate: FinDate,  # Date protection starts
+                 maturityDateOrTenor: (FinDate, str),  # FinDate or tenor
+                 runningCoupon: float,  # Annualised coupon on premium fee leg
+                 notional: float = ONE_MILLION,
+                 longProtection: bool = True,
+                 frequencyType: FinFrequencyTypes = FinFrequencyTypes.QUARTERLY,
+                 dayCountType: FinDayCountTypes = FinDayCountTypes.ACT_360,
+                 calendarType: FinCalendarTypes = FinCalendarTypes.WEEKEND,
+                 busDayAdjustType: FinBusDayAdjustTypes = FinBusDayAdjustTypes.FOLLOWING,
+                 dateGenRuleType: FinDateGenRuleTypes = FinDateGenRuleTypes.BACKWARD):
         ''' Create a CDS from the step-in date, maturity date and coupon '''
 
-        if type(stepInDate) is not FinDate:
-            raise ValueError(
-                "Step in date must be a FinDate")
+        checkArgumentTypes(self.__init__, locals())
 
         if type(maturityDateOrTenor) == FinDate:
             maturityDate = maturityDateOrTenor
         else:
+            # To get the next CDS date we move on by the tenor and then roll to
+            # the next CDS date after that. We do not holiday adjust it. That
+            # is handled in the schedule generation.
             maturityDate = stepInDate.addTenor(maturityDateOrTenor)
-
-        if type(runningCoupon) is not float and type(
-                runningCoupon) is not np.float64:
-            raise ValueError("Coupon is not float but is " +
-                             str(type(runningCoupon)))
+            maturityDate = maturityDate.nextCDSDate()
 
         if stepInDate > maturityDate:
-            raise ValueError("Step in date after maturity date")
-
-        if dayCountType not in FinDayCountTypes:
-            raise ValueError(
-                "Unknown Fixed Day Count Rule type " +
-                str(dayCountType))
-
-        if frequencyType not in FinFrequencyTypes:
-            raise ValueError(
-                "Unknown Fixed Frequency type " +
-                str(frequencyType))
-
-        if calendarType not in FinCalendarTypes:
-            raise ValueError("Unknown Calendar type " + str(calendarType))
-
-        if busDayAdjustType not in FinDayAdjustTypes:
-            raise ValueError(
-                "Unknown Business Day Adjust type " +
-                str(busDayAdjustType))
-
-        if dateGenRuleType not in FinDateGenRuleTypes:
-            raise ValueError(
-                "Unknown Date Gen Rule type " +
-                str(dateGenRuleType))
+            raise FinError("Step in date after maturity date")
 
         self._stepInDate = stepInDate
         self._maturityDate = maturityDate
-        self._coupon = runningCoupon
+        self._runningCoupon = runningCoupon
         self._notional = notional
         self._longProtection = longProtection
         self._dayCountType = dayCountType
@@ -252,12 +225,12 @@ class FinCDS(object):
         self._frequencyType = frequencyType
         self._busDayAdjustType = busDayAdjustType
 
-        self.generateAdjustedCDSPaymentDates()
-        self.calcFlows()
+        self._generateAdjustedCDSPaymentDates()
+        self._calcFlows()
 
-##########################################################################
+###############################################################################
 
-    def generateAdjustedCDSPaymentDates(self):
+    def _generateAdjustedCDSPaymentDates(self):
         ''' Generate CDS payment dates which have been holiday adjusted.'''
         frequency = FinFrequency(self._frequencyType)
         calendar = FinCalendar(self._calendarType)
@@ -325,12 +298,12 @@ class FinCDS(object):
             self._adjustedDates.append(finalDate)
 
         else:
-            raise ValueError("Unknown FinDateGenRuleType:" +
-                             str(self._dateGenRuleType))
+            raise FinError("Unknown FinDateGenRuleType:" +
+                           str(self._dateGenRuleType))
 
-##########################################################################
+###############################################################################
 
-    def calcFlows(self):
+    def _calcFlows(self):
         ''' Calculate cash flow amounts on premium leg. '''
         paymentDates = self._adjustedDates
         dayCount = FinDayCount(self._dayCountType)
@@ -346,13 +319,13 @@ class FinCDS(object):
         for it in range(1, numFlows):
             t0 = paymentDates[it - 1]
             t1 = paymentDates[it]
-            accrualFactor = dayCount.yearFrac(t0, t1)
-            flow = accrualFactor * self._coupon * self._notional
+            accrualFactor = dayCount.yearFrac(t0, t1)[0]
+            flow = accrualFactor * self._runningCoupon * self._notional
 
             self._accrualFactors.append(accrualFactor)
             self._flows.append(flow)
 
-##########################################################################
+###############################################################################
 
     def value(self,
               valuationDate,
@@ -385,13 +358,15 @@ class FinCDS(object):
             longProt = -1
 
         fullPV = fwdDf * longProt * \
-            (protPV - self._coupon * fullRPV01 * self._notional)
+            (protPV - self._runningCoupon * fullRPV01 * self._notional)
         cleanPV = fwdDf * longProt * \
-            (protPV - self._coupon * cleanRPV01 * self._notional)
+            (protPV - self._runningCoupon * cleanRPV01 * self._notional)
+
+#        print("protLeg", protPV, "cleanRPV01", cleanRPV01, "value", cleanPV)
 
         return {'full_pv': fullPV, 'clean_pv': cleanPV}
 
-##########################################################################
+###############################################################################
 
     def creditDV01(self,
                    valuationDate,
@@ -410,37 +385,34 @@ class FinCDS(object):
                         prot_method,
                         numStepsPerYear)
 
-        survProbs = issuerCurve._values
+        bump = 0.0001  # 1 basis point
 
-        bump = 0.0001
-        for cds in issuerCurve._cdsContracts:
-            cds._coupon += bump
-        issuerCurve.buildCurve()
+        # we create a deep copy to avoid state issues
+        bumpedIssuerCurve = deepcopy(issuerCurve)
+        for cds in bumpedIssuerCurve._cdsContracts:
+            cds._runningCoupon += bump
+
+        bumpedIssuerCurve._buildCurve()
 
         v1 = self.value(valuationDate,
-                        issuerCurve,
+                        bumpedIssuerCurve,
                         contractRecovery,
                         pv01Method,
                         prot_method,
                         numStepsPerYear)
 
-        # NEED TO UNDO CHANGES TO CURVE OBJECT - NO NEED TO REBUILD !!!
-        for cds in issuerCurve._cdsContracts:
-            cds._coupon -= bump
-        issuerCurve._values = survProbs
-
         creditDV01 = (v1['full_pv'] - v0['full_pv'])
         return creditDV01
 
-##########################################################################
+###############################################################################
 
     def interestDV01(self,
-                     valuationDate,
+                     valuationDate: FinDate,
                      issuerCurve,
                      contractRecovery=standardRecovery,
-                     pv01Method=0,
-                     prot_method=0,
-                     numStepsPerYear=25):
+                     pv01Method: int = 0,
+                     prot_method: int = 0,
+                     numStepsPerYear: int = 25):
         ''' Calculation of the interest DV01 based on a simple bump of
         the discount factors and reconstruction of the CDS curve. '''
 
@@ -451,42 +423,33 @@ class FinCDS(object):
                         prot_method,
                         numStepsPerYear)
 
-        dfs = issuerCurve._liborCurve._values
-        survProbs = issuerCurve._values
+        # we create a deep copy to avoid state issues
+        newIssuerCurve = deepcopy(issuerCurve)
 
-        bump = 0.0001
-        for depo in issuerCurve._liborCurve._usedDeposits:
+        bump = 0.0001  # 1 basis point
+
+        for depo in newIssuerCurve._liborCurve._usedDeposits:
             depo._depositRate += bump
-        for fra in issuerCurve._liborCurve._usedFRAs:
+        for fra in newIssuerCurve._liborCurve._usedFRAs:
             fra._fraRate += bump
-        for swap in issuerCurve._liborCurve._usedSwaps:
+        for swap in newIssuerCurve._liborCurve._usedSwaps:
             swap._fixedCoupon += bump
-        issuerCurve._liborCurve.buildCurve()
-        issuerCurve.buildCurve()
+
+        newIssuerCurve._liborCurve._buildCurve()
+
+        newIssuerCurve._buildCurve()
 
         v1 = self.value(valuationDate,
-                        issuerCurve,
+                        newIssuerCurve,
                         contractRecovery,
                         pv01Method,
                         prot_method,
                         numStepsPerYear)
 
-        # Need do undo changes to Libor curve object in issuer curve
-        # Do not need to rebuild it - just restore initial dfs and surv probs.
-        for depo in issuerCurve._liborCurve._usedDeposits:
-            depo._depositRate -= bump
-        for fra in issuerCurve._liborCurve._usedFRAs:
-            fra._fraRate -= bump
-        for swap in issuerCurve._liborCurve._usedSwaps:
-            swap._fixedCoupon -= bump
+        interestDV01 = (v1['full_pv'] - v0['full_pv'])
+        return interestDV01
 
-        issuerCurve._liborCurve._values = dfs
-        issuerCurve._values = survProbs
-
-        creditDV01 = (v1[0] - v0[0])
-        return creditDV01
-
-##########################################################################
+###############################################################################
 
     def cashSettlementAmount(self,
                              valuationDate,
@@ -511,7 +474,7 @@ class FinCDS(object):
         v = v / df
         return v
 
-##########################################################################
+###############################################################################
 
     def cleanPrice(self,
                    valuationDate,
@@ -534,11 +497,12 @@ class FinCDS(object):
 
         fwdDf = 1.0
 
-        cleanPV = fwdDf * (protPV - self._coupon * cleanRPV01 * self._notional)
+        cleanPV = fwdDf * (protPV - self._runningCoupon * cleanRPV01
+                           * self._notional)
         cleanPrice = (self._notional - cleanPV) / self._notional * 100.0
         return cleanPrice
 
-##########################################################################
+###############################################################################
 
     def riskyPV01_OLD(self,
                       valuationDate,
@@ -567,10 +531,10 @@ class FinCDS(object):
 
         # this is the part of the coupon accrued from the previous coupon date
         # to now
-        accrualFactorPCDToNow = dayCount.yearFrac(pcd, teff)
+        accrualFactorPCDToNow = dayCount.yearFrac(pcd, teff)[0]
 
         # full first coupon is paid at the end of the current period if the
-        yearFrac = dayCount.yearFrac(pcd, ncd)
+        yearFrac = dayCount.yearFrac(pcd, ncd)[0]
 
         # reference credit survives to the premium payment date
         fullRPV01 = q1 * z1 * yearFrac
@@ -583,7 +547,8 @@ class FinCDS(object):
         # future accrued from now to coupon payment date assuming default
         # roughly midway
         fullRPV01 = fullRPV01 + 0.5 * z1 * \
-            (qeff - q1) * (yearFrac - accrualFactorPCDToNow) * couponAccruedIndicator
+            (qeff - q1) * (yearFrac - accrualFactorPCDToNow) \
+            * couponAccruedIndicator
 
         for it in range(2, len(paymentDates)):
 
@@ -592,7 +557,7 @@ class FinCDS(object):
             q2 = issuerCurve.survivalProbability(t2)
             z2 = issuerCurve.df(t2)
 
-            accrualFactor = dayCount.yearFrac(t1, t2)
+            accrualFactor = dayCount.yearFrac(t1, t2)[0]
 
             # full coupon is paid at the end of the current period if survives
             # to payment date
@@ -627,7 +592,7 @@ class FinCDS(object):
 
         return {'full_rpv01': fullRPV01, 'clean_rpv01': cleanRPV01}
 
-##########################################################################
+###############################################################################
 
     def accruedDays(self):
         ''' Number of days between the previous coupon and the currrent step
@@ -639,7 +604,7 @@ class FinCDS(object):
         accruedDays = (self._stepInDate - pcd)
         return accruedDays
 
-##########################################################################
+###############################################################################
 
     def accruedInterest(self):
         ''' Calculate the amount of accrued interest that has accrued from the
@@ -648,15 +613,15 @@ class FinCDS(object):
         dayCount = FinDayCount(self._dayCountType)
         paymentDates = self._adjustedDates
         pcd = paymentDates[0]
-        accrualFactor = dayCount.yearFrac(pcd, self._stepInDate)
-        accruedInterest = accrualFactor * self._notional * self._coupon
+        accrualFactor = dayCount.yearFrac(pcd, self._stepInDate)[0]
+        accruedInterest = accrualFactor * self._notional * self._runningCoupon
 
         if self._longProtection:
             accruedInterest *= -1.0
 
         return accruedInterest
 
-##########################################################################
+###############################################################################
 
     def protectionLegPV(self,
                         valuationDate,
@@ -672,19 +637,19 @@ class FinCDS(object):
 
         liborCurve = issuerCurve._liborCurve
 
-        v = protectionLegPV_NUMBA(teff,
-                                  tmat,
-                                  liborCurve._times,
-                                  liborCurve._values,
-                                  issuerCurve._times,
-                                  issuerCurve._values,
-                                  contractRecovery,
-                                  numStepsPerYear,
-                                  protMethod)
+        v = _protectionLegPV_NUMBA(teff,
+                                   tmat,
+                                   liborCurve._times,
+                                   liborCurve._dfValues,
+                                   issuerCurve._times,
+                                   issuerCurve._values,
+                                   contractRecovery,
+                                   numStepsPerYear,
+                                   protMethod)
 
         return v * self._notional
 
-##########################################################################
+###############################################################################
 
     def riskyPV01(self,
                   valuationDate,
@@ -706,20 +671,20 @@ class FinCDS(object):
         eff = self._stepInDate
         dayCount = FinDayCount(self._dayCountType)
 
-        accrualFactorPCDToNow = dayCount.yearFrac(pcd, eff)
+        accrualFactorPCDToNow = dayCount.yearFrac(pcd, eff)[0]
 
         yearFracs = self._accrualFactors
         teff = (eff - valuationDate) / gDaysInYear
 
-        valueRPV01 = riskyPV01_NUMBA(teff,
-                                     accrualFactorPCDToNow,
-                                     np.array(paymentTimes),
-                                     np.array(yearFracs),
-                                     liborCurve._times,
-                                     liborCurve._values,
-                                     issuerCurve._times,
-                                     issuerCurve._values,
-                                     pv01Method)
+        valueRPV01 = _riskyPV01_NUMBA(teff,
+                                      accrualFactorPCDToNow,
+                                      np.array(paymentTimes),
+                                      np.array(yearFracs),
+                                      liborCurve._times,
+                                      liborCurve._dfValues,
+                                      issuerCurve._times,
+                                      issuerCurve._values,
+                                      pv01Method)
 
         fullRPV01 = valueRPV01[0]
         cleanRPV01 = valueRPV01[1]
@@ -727,7 +692,7 @@ class FinCDS(object):
 #        print("NEW PV01",fullRPV01, cleanRPV01)
         return {'full_rpv01': fullRPV01, 'clean_rpv01': cleanRPV01}
 
-##########################################################################
+###############################################################################
 
     def premiumLegPV(self,
                      valuationDate,
@@ -739,10 +704,10 @@ class FinCDS(object):
                                    issuerCurve,
                                    pv01Method)['full_rpv01']
 
-        v = fullRPV01 * self._notional * self._coupon
+        v = fullRPV01 * self._notional * self._runningCoupon
         return v
 
-##########################################################################
+###############################################################################
 
     def parSpread(self,
                   valuationDate,
@@ -768,7 +733,7 @@ class FinCDS(object):
         spd = prot / cleanRPV01 / self._notional
         return spd
 
-##########################################################################
+###############################################################################
 
     def valueFastApprox(self,
                         valuationDate,
@@ -780,9 +745,8 @@ class FinCDS(object):
         accurate approximation that avoids curve building. '''
 
         if type(valuationDate) is not FinDate:
-            raise ValueError(
-                "Valuation date must be a FinDate and not " +
-                str(valuationDate))
+            raise FinError("Valuation date must be a FinDate and not " +
+                           str(valuationDate))
 
         t_mat = (self._maturityDate - valuationDate) / gDaysInYear
         t_eff = (self._stepInDate - valuationDate) / gDaysInYear
@@ -790,80 +754,63 @@ class FinCDS(object):
         h = flatCDSCurveSpread / (1.0 - curveRecovery)
         r = flatContinuousInterestRate
         fwdDf = 1.0
+        bumpSize = 0.0001
 
         if self._longProtection:
             longProtection = +1
         else:
             longProtection = -1
 
+        # The sign of he accrued has already been sign adjusted for direction
         accrued = self.accruedInterest()
 
         # This is the clean RPV01 as it treats the PV01 stream as though it
         # pays just the accrued for the time between 0 and the maturity
         # It therefore omits the part that has accrued
-        cleanRPV01 = (exp(-(r + h) * t_eff) - exp(-(r + h)
-                                                  * t_mat)) / (h + r) * 365.0 / 360.0
-        protPV = h * (1.0 - contractRecovery) * (exp(-(r + h) * \
-                      t_eff) - exp(-(r + h) * t_mat)) / (r + h) * self._notional
-        cleanPV = fwdDf * longProtection * \
-            (protPV - self._coupon * cleanRPV01 * self._notional)
-        fullPV = cleanPV + fwdDf * longProtection * accrued
 
-        bumpSize = 0.0001
+        w = r + h
+        z = np.exp(-w * t_eff) - np.exp(-w * t_mat)
+        cleanRPV01 = (z / w) * 365.0 / 360.0
+        protPV = h * (1.0 - contractRecovery) * (z / w) * self._notional
+        cleanPV = fwdDf * longProtection * \
+            (protPV - self._runningCoupon * cleanRPV01 * self._notional)
+        fullPV = cleanPV + fwdDf * accrued
+
+        #######################################################################
+        # bump CDS spread and calculate
+        #######################################################################
 
         h = (flatCDSCurveSpread + bumpSize) / (1.0 - contractRecovery)
         r = flatContinuousInterestRate
-        cleanRPV01 = (exp(-(r + h) * t_eff) - exp(-(r + h)
-                                                  * t_mat)) / (h + r) * 365.0 / 360.0
-        protPV = h * (1.0 - contractRecovery) * (exp(-(r + h) * \
-                      t_eff) - exp(-(r + h) * t_mat)) / (r + h) * self._notional
+        w = r + h
+        z = np.exp(-w * t_eff) - np.exp(-w * t_mat)
+        cleanRPV01 = (z / w) * 365.0 / 360.0
+        protPV = h * (1.0 - contractRecovery) * (z / w) * self._notional
         cleanPV_credit_bumped = fwdDf * longProtection * \
-            (protPV - self._coupon * cleanRPV01 * self._notional)
-        fullPV_credit_bumped = cleanPV_credit_bumped + fwdDf * longProtection * accrued
+            (protPV - self._runningCoupon * cleanRPV01 * self._notional)
+        fullPV_credit_bumped = cleanPV_credit_bumped \
+            + fwdDf * longProtection * accrued
         credit01 = fullPV_credit_bumped - fullPV
 
+        #######################################################################
+        # bump Rate and calculate
+        #######################################################################
+
         h = flatCDSCurveSpread / (1.0 - contractRecovery)
         r = flatContinuousInterestRate + bumpSize
-        cleanRPV01 = (exp(-(r + h) * t_eff) - exp(-(r + h)
-                                                  * t_mat)) / (h + r) * 365.0 / 360.0
-        protPV = h * (1.0 - contractRecovery) * (exp(-(r + h) * \
-                      t_eff) - exp(-(r + h) * t_mat)) / (r + h) * self._notional
+
+        w = r + h
+        z = np.exp(-w * t_eff) - np.exp(-w * t_mat)
+        cleanRPV01 = (z / w) * 365.0 / 360.0
+        protPV = h * (1.0 - contractRecovery) * (z/w) * self._notional
         cleanPV_ir_bumped = fwdDf * longProtection * \
-            (protPV - self._coupon * cleanRPV01 * self._notional)
+            (protPV - self._runningCoupon * cleanRPV01 * self._notional)
         fullPV_ir_bumped = cleanPV_ir_bumped + fwdDf * longProtection * accrued
         ir01 = fullPV_ir_bumped - fullPV
 
         return (fullPV, cleanPV, credit01, ir01)
 
-##########################################################################
-
-    def print(self, valuationDate):
-        ''' print out details of the CDS contract and all of the calculated
-        cashflows '''
-        print("STEPINDATE: ", str(self._stepInDate))
-        print("MATURITY: ", str(self._maturityDate))
-        print("NOTIONAL:", str(self._notional))
-        print("RUNNING COUPON: ", str(self._coupon * 10000), "bp")
-        print("DAYCOUNT: ", str(self._dayCountType))
-        print("FREQUENCY: ", str(self._frequencyType))
-        print("CALENDAR: ", str(self._calendarType))
-        print("BUSDAYRULE: ", str(self._busDayAdjustType))
-        print("DATEGENRULE: ", str(self._dateGenRuleType))
-
-        accruedDays = self.accruedDays()
-        print("ACCRUED DAYS:", str(accruedDays))
-
-        numFlows = len(self._adjustedDates)
-
-        print("PAYMENT_DATE      YEAR_FRAC      FLOW")
-
-        for it in range(1, numFlows):
-            dt = self._adjustedDates[it]
-            accFactor = self._accrualFactors[it]
-            flow = self._flows[it]
-            print("%15s %10.6f %12.2f" % (dt, accFactor, flow))
-
-##########################################################################
+###############################################################################
 
     def printFlows(self, issuerCurve):
 
@@ -880,4 +827,35 @@ class FinCDS(object):
             print("%15s %10.6f %12.2f %12.6f %12.6f %12.2f" %
                   (dt, accFactor, flow, z, q, flow * z * q))
 
-##########################################################################
+###############################################################################
+
+    def __repr__(self):
+        ''' print out details of the CDS contract and all of the calculated
+        cashflows '''
+        s = labelToString("OBJECT TYPE", type(self).__name__)
+        s += labelToString("STEP-IN DATE", self._stepInDate)
+        s += labelToString("MATURITY", self._maturityDate)
+        s += labelToString("NOTIONAL", self._notional)
+        s += labelToString("RUNNING COUPON", self._runningCoupon*10000, "bp\n")
+        s += labelToString("DAYCOUNT", self._dayCountType)
+        s += labelToString("FREQUENCY", self._frequencyType)
+        s += labelToString("CALENDAR", self._calendarType)
+        s += labelToString("BUSDAYRULE", self._busDayAdjustType)
+        s += labelToString("DATEGENRULE", self._dateGenRuleType)
+        s += labelToString("ACCRUED DAYS", self.accruedDays())
+
+        header = "PAYMENT_DATE, YEAR_FRAC, FLOW"
+        valueTable = [self._adjustedDates, self._accrualFactors, self._flows]
+        precision = "12.6f"
+
+        s += tableToString(header, valueTable, precision)
+
+        return s
+
+###############################################################################
+
+    def _print(self):
+        ''' Simple print function for backward compatibility. '''
+        print(self)
+
+###############################################################################
