@@ -1,19 +1,24 @@
-# -*- coding: utf-8 -*-
-"""
-Created on Wed Jan 16 17:54:03 2019
+##############################################################################
+# Copyright (C) 2018, 2019, 2020 Dominic O'Kane
+##############################################################################
 
-@author: Dominic O'Kane
-"""
 
 from math import pow
 
 from ...finutils.FinCalendar import FinCalendarTypes
-from ...finutils.FinCalendar import FinDayAdjustTypes, FinDateGenRuleTypes
+from ...finutils.FinCalendar import FinBusDayAdjustTypes, FinDateGenRuleTypes
 from ...finutils.FinDayCount import FinDayCountTypes
 from ...finutils.FinFrequency import FinFrequencyTypes
 from ...finutils.FinError import FinError
 from ...products.credit.FinCDS import FinCDS
-from ...market.curves.FinCDSCurve import FinCDSCurve
+from ...products.credit.FinCDSCurve import FinCDSCurve
+from ...finutils.FinHelperFunctions import checkArgumentTypes
+from ...finutils.FinHelperFunctions import labelToString
+
+###############################################################################
+# TODO: Move index spread details into class and then pass in issuer curves
+#       to the function when doing the adjustment
+###############################################################################
 
 
 class FinCDSIndexPortfolio():
@@ -21,19 +26,21 @@ class FinCDSIndexPortfolio():
     portfolio of CDS contracts with the same maturity date. '''
 
     def __init__(self,
-                 frequencyType=FinFrequencyTypes.QUARTERLY,
-                 dayCountType=FinDayCountTypes.ACT_360,
-                 calendarType=FinCalendarTypes.WEEKEND,
-                 busDayAdjustType=FinDayAdjustTypes.FOLLOWING,
-                 dateGenRuleType=FinDateGenRuleTypes.BACKWARD):
+                 frequencyType: FinFrequencyTypes = FinFrequencyTypes.QUARTERLY,
+                 dayCountType: FinDayCountTypes = FinDayCountTypes.ACT_360,
+                 calendarType: FinCalendarTypes = FinCalendarTypes.WEEKEND,
+                 busDayAdjustType: FinBusDayAdjustTypes = FinBusDayAdjustTypes.FOLLOWING,
+                 dateGenRuleType: FinDateGenRuleTypes = FinDateGenRuleTypes.BACKWARD):
         ''' Create FinCDSIndexPortfolio object. Note that all of the inputs
         have a default value which reflects the CDS market standard. '''
 
+        checkArgumentTypes(self.__init__, locals())
+
         self._dayCountType = dayCountType
         self._dateGenRuleType = dateGenRuleType
         self._calendarType = calendarType
         self._frequencyType = frequencyType
-        self._businessDateAdjustType = busDayAdjustType
+        self._busDayAdjustType = busDayAdjustType
 
 ###############################################################################
 
@@ -181,9 +188,7 @@ class FinCDSIndexPortfolio():
         numCredits = len(issuerCurves)
 
         if numCredits < 1:
-            raise FinError(
-                "Number of credits in index must be > 1 and not ",
-                str(numCredits))
+            raise FinError("Number of credits in index must be > 1 and not" + str(numCredits))
 
         cdsContract = FinCDS(stepInDate,
                              maturityDate,
@@ -211,9 +216,7 @@ class FinCDSIndexPortfolio():
         numCredits = len(issuerCurves)
 
         if numCredits < 1:
-            raise FinError(
-                "Number of credits in index must be > 1 and not ",
-                str(numCredits))
+            raise FinError("Number of credits in index must be > 1 and not " + str(numCredits))
 
         cdsContract = FinCDS(stepInDate,
                              maturityDate,
@@ -230,7 +233,8 @@ class FinCDSIndexPortfolio():
 
 ###############################################################################
 
-    def spreadAdjustIntrinsic(valuationDate,
+    def spreadAdjustIntrinsic(self,
+                              valuationDate,
                               issuerCurves,
                               indexCoupons,
                               indexUpfronts,
@@ -245,9 +249,7 @@ class FinCDSIndexPortfolio():
         numCredits = len(issuerCurves)
 
         if numCredits < 1:
-            raise FinError(
-                "Number of credits in index must be > 1 and not ",
-                str(numCredits))
+            raise FinError("Number of credits in index must be > 1 and not " + str(numCredits))
 
         liborCurve = issuerCurves[0]._liborCurve
         numIndexMaturityPoints = len(indexCoupons)
@@ -319,10 +321,10 @@ class FinCDSIndexPortfolio():
 
                     for j in range(0, numCDSMaturityPoints):
 
-                        cdsSpread = cdsContracts[j]._coupon
+                        cdsSpread = cdsContracts[j]._runningCoupon
                         adjustedCDSSpreads[j] = cdsSpread * \
                             cdsSpreadMultipliers[j]
-                        curveCDSContracts[j]._coupon = adjustedCDSSpreads[j]
+                        curveCDSContracts[j]._runningCoupon = adjustedCDSSpreads[j]
 
                     adjustedIssuerCurve = FinCDSCurve(valuationDate,
                                                       curveCDSContracts,
@@ -330,10 +332,10 @@ class FinCDSIndexPortfolio():
                                                       recoveryRate)
 
                     indexProtectionPV = cdsIndex.protectionLegPV(valuationDate,
-                                                    adjustedIssuerCurve,
-                                                    indexRecoveryRate)
+                                                                 adjustedIssuerCurve,
+                                                                 indexRecoveryRate)
 
-                    cleanRPV01 = cdsIndex.riskyPV01(valuationDate, 
+                    cleanRPV01 = cdsIndex.riskyPV01(valuationDate,
                                                     adjustedIssuerCurve)['clean_rpv01']
 
                     sumRPV01 += cleanRPV01
@@ -362,7 +364,7 @@ class FinCDSIndexPortfolio():
 
             for j in range(0, numCDSMaturityPoints):
 
-                unadjustedSpread = issuerCurves[iCredit]._cdsContracts[j]._coupon
+                unadjustedSpread = issuerCurves[iCredit]._cdsContracts[j]._runningCoupon
 
                 adjustedSpread = unadjustedSpread * cdsSpreadMultipliers[j]
 
@@ -384,7 +386,8 @@ class FinCDSIndexPortfolio():
 
 ###############################################################################
 
-    def hazardRateAdjustIntrinsic(valuationDate,
+    def hazardRateAdjustIntrinsic(self,
+                                  valuationDate,
                                   issuerCurves,
                                   indexCoupons,
                                   indexUpfronts,
@@ -409,7 +412,7 @@ class FinCDSIndexPortfolio():
         for issuerCurve in issuerCurves:
 
             adjustedIssuerCurve = FinCDSCurve(valuationDate,
-                                              None,
+                                              [],
                                               liborCurve)
 
             adjustedIssuerCurve._times = issuerCurve._times.copy()
@@ -447,11 +450,11 @@ class FinCDSIndexPortfolio():
 
                     # the CDS spreads we extract here should be the index
                     # maturity dates
-                    cdsIndex = FinCDS(
-                        valuationDate, indexMaturityDate, 0.0, 1.0)
+                    cdsIndex = FinCDS(valuationDate, indexMaturityDate, 0, 1.0)
 
-                    indexProtPV = cdsIndex.protectionLegPV(
-                        valuationDate, adjustedIssuerCurves[iCredit], indexRecoveryRate)
+                    indexProtPV = cdsIndex.protectionLegPV(valuationDate,
+                                                           adjustedIssuerCurves[iCredit],
+                                                           indexRecoveryRate)
 
                     rpv01Ret = cdsIndex.riskyPV01(
                         valuationDate, adjustedIssuerCurves[iCredit])
@@ -473,4 +476,22 @@ class FinCDSIndexPortfolio():
 
         return adjustedIssuerCurves
 
+###############################################################################
+
+    def __repr__(self):
+
+        s = labelToString("OBJECT TYPE", type(self).__name__)
+        s += labelToString("FREQUENCY", self._frequencyType)
+        s += labelToString("DAYCOUNT", self._dayCountType)
+        s += labelToString("CALENDAR", self._calendarType)
+        s += labelToString("BUSDAYRULE", self._busDayAdjustType)
+        s += labelToString("DATEGENRULE", self._dateGenRuleType)
+        return s
+
+###############################################################################
+
+    def _print(self):
+        ''' Simple print function for backward compatibility. '''
+        print(self)
+
 ###############################################################################
