@@ -1,22 +1,37 @@
-# -*- coding: utf-8 -*-
-"""
-Created on Fri Feb 12 16:51:05 2016
+##############################################################################
+# Copyright (C) 2018, 2019, 2020 Dominic O'Kane
+##############################################################################
 
-@author: Dominic O'Kane
-"""
 
 import numpy as np
-
-from math import exp, log, sqrt
-
 from numba import njit
 
+# TODO: Add perturbatory risk using the analytical methods !!
+# TODO: Add Sobol to Monte Carlo
+
 from ...finutils.FinMath import N, covar
 from ...finutils.FinGlobalVariables import gDaysInYear
 from ...finutils.FinError import FinError
 
-from ...products.equity.FinEquityOption import FinEquityOption
-from ...products.FinOptionTypes import FinOptionTypes
+from ...finutils.FinOptionTypes import FinOptionTypes
+from ...finutils.FinHelperFunctions import checkArgumentTypes, labelToString
+from ...finutils.FinDate import FinDate
+from ...market.curves.FinDiscountCurve import FinDiscountCurve
+
+###############################################################################
+
+from enum import Enum
+
+
+class FinAsianOptionValuationMethods(Enum):
+    GEOMETRIC = 1,
+    TURNBULL_WAKEMAN = 2,
+    CURRAN = 3
+
+###############################################################################
+
+
+errorStr = "In averaging period so need to enter accrued average."
 
 ###############################################################################
 # An Asian option on an arithmetic average and strike K has a payoff
@@ -48,25 +63,28 @@ from ...products.FinOptionTypes import FinOptionTypes
 
 
 @njit(cache=True, fastmath=True)
-def valueMC_NUMBA(t0, t, tau, K, n, optionType,
-                  stockPrice,
-                  interestRate,
-                  dividendYield,
-                  volatility,
-                  numPaths,
-                  seed,
-                  accruedAverage):
+def _valueMC_NUMBA(t0,
+                   t,
+                   tau,
+                   K,
+                   n,
+                   optionType,
+                   stockPrice,
+                   interestRate,
+                   dividendYield,
+                   volatility,
+                   numPaths,
+                   seed,
+                   accruedAverage):
 
-    # Start pricing
+    # Start pricing here
     np.random.seed(seed)
-
     multiplier = 1.0
 
-    if t0 < 0:  # we are in the averaging period
+    if t0 < 0.0:  # we are in the averaging period
 
         if accruedAverage is None:
-            raise FinError(
-            "In the averaging period you need to enter the accrued average")
+            raise FinError(errorStr)
 
         # we adjust the strike to account for the accrued coupon
         K = (K * tau + accruedAverage * t0) / t
@@ -83,14 +101,14 @@ def valueMC_NUMBA(t0, t, tau, K, n, optionType,
 
     payoff_a = 0.0
 
-    for p in range(0, numPaths):
+    for _ in range(0, numPaths):
 
         # evolve stock price to start of averaging period
         g = np.random.standard_normal(1)
-        s_1 = stockPrice * exp((mu - v2 / 2.0) * t0 +
-                               g[0] * sqrt(t0) * volatility)
-        s_2 = stockPrice * exp((mu - v2 / 2.0) * t0 -
-                               g[0] * sqrt(t0) * volatility)
+        s_1 = stockPrice * np.exp((mu - v2 / 2.0) * t0 +
+                                  g[0] * np.sqrt(t0) * volatility)
+        s_2 = stockPrice * np.exp((mu - v2 / 2.0) * t0 -
+                                  g[0] * np.sqrt(t0) * volatility)
 
         # enter averaging period
         s_1_arithmetic = 0.0
@@ -100,10 +118,10 @@ def valueMC_NUMBA(t0, t, tau, K, n, optionType,
 
         for obs in range(0, n):
 
-            s_1 = s_1 * exp((mu - v2 / 2.0) * dt +
-                            g[obs] * sqrt(dt) * volatility)
-            s_2 = s_2 * exp((mu - v2 / 2.0) * dt -
-                            g[obs] * sqrt(dt) * volatility)
+            s_1 = s_1 * np.exp((mu - v2 / 2.0) * dt +
+                               g[obs] * np.sqrt(dt) * volatility)
+            s_2 = s_2 * np.exp((mu - v2 / 2.0) * dt -
+                               g[obs] * np.sqrt(dt) * volatility)
 
             s_1_arithmetic += s_1
             s_2_arithmetic += s_2
@@ -120,7 +138,7 @@ def valueMC_NUMBA(t0, t, tau, K, n, optionType,
         else:
             return None
 
-    v_a = payoff_a * exp(-interestRate * t) / numPaths / 2.0
+    v_a = payoff_a * np.exp(-interestRate * t) / numPaths / 2.0
     v_a = v_a * multiplier
     return v_a
 
@@ -128,29 +146,33 @@ def valueMC_NUMBA(t0, t, tau, K, n, optionType,
 
 
 @njit(cache=True, fastmath=True)
-def valueMC_fast_NUMBA(t0, t, tau, K, n, optionType,
-                       stockPrice,
-                       interestRate,
-                       dividendYield,
-                       volatility,
-                       numPaths,
-                       seed,
-                       accruedAverage):
+def _valueMC_fast_NUMBA(t0: float,
+                        t: float,
+                        tau: float,
+                        K: float,
+                        n: int,
+                        optionType: FinOptionTypes,
+                        stockPrice: float,
+                        interestRate: float,
+                        dividendYield: float,
+                        volatility: float,
+                        numPaths: int,
+                        seed: int,
+                        accruedAverage: float):
 
     np.random.seed(seed)
-
     mu = interestRate - dividendYield
     v2 = volatility**2
     dt = (t - t0) / n
     r = interestRate
+    numPaths = int(numPaths)
 
     multiplier = 1.0
 
-    if t0 < 0:  # we are in the averaging period
+    if t0 < 0.0:  # we are in the averaging period
 
         if accruedAverage is None:
-            raise FinError(
-                "In the averaging period you need to enter the accrued average")
+            raise FinError(errorStr)
 
         # we adjust the strike to account for the accrued coupon
         K = (K * tau + accruedAverage * t0) / t
@@ -161,32 +183,35 @@ def valueMC_fast_NUMBA(t0, t, tau, K, n, optionType,
         # the number of observations is scaled and floored at 1
         n = int(n * t / tau + 0.5) + 1
 
-#    payoff_a = 0.0
-
     # evolve stock price to start of averaging period
-    g = np.random.normal(0.0, 1.0, size=(numPaths))
+    # g = np.random.normal(0.0, 1.0, size=(numPaths))
+
+    gg = np.empty(numPaths, np.float64)
+    for iPath in range(0, numPaths):
+        rv = np.random.normal()
+        gg[iPath] = rv
 
     s_1 = np.empty(numPaths)
     s_2 = np.empty(numPaths)
 
     for ip in range(0, numPaths):
         s_1[ip] = stockPrice * \
-            exp((mu - v2 / 2.0) * t0 + g[ip] * sqrt(t0) * volatility)
+            np.exp((mu - v2 / 2.0) * t0 + gg[ip] * np.sqrt(t0) * volatility)
         s_2[ip] = stockPrice * \
-            exp((mu - v2 / 2.0) * t0 - g[ip] * sqrt(t0) * volatility)
+            np.exp((mu - v2 / 2.0) * t0 - gg[ip] * np.sqrt(t0) * volatility)
 
     s_1_arithmetic = np.zeros(numPaths)
     s_2_arithmetic = np.zeros(numPaths)
 
-    for obs in range(0, n):
+    for _ in range(0, n):
 
         g = np.random.normal(0.0, 1.0, size=(numPaths))
 
         for ip in range(0, numPaths):
-            s_1[ip] = s_1[ip] * exp((mu - v2 / 2.0) *
-                                    dt + g[ip] * sqrt(dt) * volatility)
-            s_2[ip] = s_2[ip] * exp((mu - v2 / 2.0) *
-                                    dt - g[ip] * sqrt(dt) * volatility)
+            s_1[ip] = s_1[ip] * np.exp((mu - v2 / 2.0) *
+                                       dt + g[ip] * np.sqrt(dt) * volatility)
+            s_2[ip] = s_2[ip] * np.exp((mu - v2 / 2.0) *
+                                       dt - g[ip] * np.sqrt(dt) * volatility)
 
         for ip in range(0, numPaths):
             s_1_arithmetic[ip] += s_1[ip] / n
@@ -202,25 +227,18 @@ def valueMC_fast_NUMBA(t0, t, tau, K, n, optionType,
         raise FinError("Unknown option type.")
 
     payoff_a = np.mean(payoff_a_1) + np.mean(payoff_a_2)
-    v_a = multiplier * payoff_a * exp(- r * t) / 2.0
+    v_a = multiplier * payoff_a * np.exp(- r * t) / 2.0
     return v_a
 
-##########################################################################
+###############################################################################
 
 
 @njit(cache=True, fastmath=True)
-def valueMC_fast_CV_NUMBA(t0, t, tau, K, n, optionType,
-                          stockPrice,
-                          interestRate,
-                          dividendYield,
-                          volatility,
-                          numPaths,
-                          seed,
-                          accruedAverage,
-                          v_g_exact):
+def _valueMC_fast_CV_NUMBA(t0, t, tau, K, n, optionType, stockPrice,
+                           interestRate, dividendYield, volatility, numPaths,
+                           seed, accruedAverage, v_g_exact):
 
     np.random.seed(seed)
-
     mu = interestRate - dividendYield
     v2 = volatility**2
     dt = (t - t0) / n
@@ -231,8 +249,7 @@ def valueMC_fast_CV_NUMBA(t0, t, tau, K, n, optionType,
     if t0 < 0:  # we are in the averaging period
 
         if accruedAverage is None:
-            raise FinError(
-                "In averaging period you need to enter the accrued average")
+            raise FinError(errorStr)
 
         # we adjust the strike to account for the accrued coupon
         K = (K * tau + accruedAverage * t0) / t
@@ -251,23 +268,23 @@ def valueMC_fast_CV_NUMBA(t0, t, tau, K, n, optionType,
 
     for ip in range(0, numPaths):
         s_1[ip] = stockPrice * \
-            exp((mu - v2 / 2.0) * t0 + g[ip] * sqrt(t0) * volatility)
+            np.exp((mu - v2 / 2.0) * t0 + g[ip] * np.sqrt(t0) * volatility)
         s_2[ip] = stockPrice * \
-            exp((mu - v2 / 2.0) * t0 - g[ip] * sqrt(t0) * volatility)
+            np.exp((mu - v2 / 2.0) * t0 - g[ip] * np.sqrt(t0) * volatility)
 
     s_1_arithmetic = np.zeros(numPaths)
     s_2_arithmetic = np.zeros(numPaths)
     ln_s_1_geometric = np.zeros(numPaths)
     ln_s_2_geometric = np.zeros(numPaths)
 
-    for obs in range(0, n):
+    for _ in range(0, n):
 
         g = np.random.normal(0.0, 1.0, size=(numPaths))
         for ip in range(0, numPaths):
-            s_1[ip] = s_1[ip] * exp((mu - v2 / 2.0) *
-                                    dt + g[ip] * sqrt(dt) * volatility)
-            s_2[ip] = s_2[ip] * exp((mu - v2 / 2.0) *
-                                    dt - g[ip] * sqrt(dt) * volatility)
+            s_1[ip] = s_1[ip] * np.exp((mu - v2 / 2.0) *
+                                       dt + g[ip] * np.sqrt(dt) * volatility)
+            s_2[ip] = s_2[ip] * np.exp((mu - v2 / 2.0) *
+                                       dt - g[ip] * np.sqrt(dt) * volatility)
 
         for ip in range(0, numPaths):
             s_1_arithmetic[ip] += s_1[ip]
@@ -300,32 +317,43 @@ def valueMC_fast_CV_NUMBA(t0, t, tau, K, n, optionType,
     payoff_a = np.concatenate((payoff_a_1, payoff_a_2), axis=0)
     payoff_g = np.concatenate((payoff_g_1, payoff_g_2), axis=0)
 
+    # Now we do the control variate adjustment
     m = covar(payoff_a, payoff_a)
     lam = m[0][1] / m[1][1]
 
     payoff_a_mean = np.mean(payoff_a)
     payoff_g_mean = np.mean(payoff_g)
 
-    v_a = payoff_a_mean * exp(-r * t) * multiplier
-    v_g = payoff_g_mean * exp(-r * t) * multiplier
+    v_a = payoff_a_mean * np.exp(-r * t) * multiplier
+    v_g = payoff_g_mean * np.exp(-r * t) * multiplier
 
     epsilon = v_g_exact - v_g
     v_a_cv = v_a + lam * epsilon
 
     return v_a_cv
 
-##########################################################################
+###############################################################################
 
 
-class FinEquityAsianOption(FinEquityOption):
-    ''' Class to store an Equity Asian Option. '''
+class FinEquityAsianOption():
+    ''' Class for an Equity Asian Option. This is an option with a final payoff
+    linked to the averaging of the stock price over some specified period
+    before the option expires. The valuation is done for both an arithmetic and
+    a geometric average but the former can only be done either using an
+    analytical approximation of the arithmetic average distribution or by using
+    Monte-Carlo simulation.'''
 
     def __init__(self,
-                 startAveragingDate,
-                 expiryDate,
-                 strikePrice,
-                 optionType,
-                 numberOfObservations=0):
+                 startAveragingDate: FinDate,
+                 expiryDate: FinDate,
+                 strikePrice: float,
+                 optionType: FinOptionTypes,
+                 numberOfObservations: int = 100):
+        ''' Create an FinEquityAsian option object which takes a start date for
+        the averaging, an expiry date, a strike price, an option type and a
+        number of observations. '''
+
+        checkArgumentTypes(self.__init__, locals())
 
         if startAveragingDate > expiryDate:
             raise FinError("Averaging starts after expiry date")
@@ -339,63 +367,53 @@ class FinEquityAsianOption(FinEquityOption):
 ###############################################################################
 
     def value(self,
-              valueDate,
-              stockPrice,
-              discountCurve,
-              dividendYield,
+              valueDate: FinDate,
+              stockPrice: float,
+              discountCurve: FinDiscountCurve,
+              dividendYield: float,
               model,
-              valuationMethod,
-              accruedAverage=None):
-        ''' Calculate the value of an Asian option. '''
+              method: FinAsianOptionValuationMethods,
+              accruedAverage: float = None):
+        ''' Calculate the value of an Asian option using one of the specified
+        analytical approximations for an average rate option. These are the
+        three enumerated values in the enum FinAsianOptionValuationMethods. The
+        choices of approximation are (i) GEOMETRIC - the average is a geometric
+        one as in paper by Kenna and Worst (1990), (ii) TURNBULL_WAKEMAN -
+        this is a value based on an edgeworth expansion of the moments of the
+        arithmetic average, and (iii) CURRAN - another approximative approach
+        by Curran based on conditioning on the geometric mean price. Just
+        choose the corresponding enumerated value to switch between these
+        different approaches.
+
+        Note that the accrued average is only required if the value date is
+        inside the averaging period for the option. '''
 
         if valueDate > self._expiryDate:
             raise FinError("Value date after expiry date.")
 
-        if valuationMethod == "GEOMETRIC":
+        if method == FinAsianOptionValuationMethods.GEOMETRIC:
+            v = self._valueGeometric(valueDate,
+                                     stockPrice,
+                                     discountCurve,
+                                     dividendYield,
+                                     model,
+                                     accruedAverage)
 
-            v = self.valueGeometric(valueDate,
-                                    stockPrice,
-                                    discountCurve,
-                                    dividendYield,
-                                    model,
-                                    accruedAverage)
-
-        elif valuationMethod == "TURNBULL_WAKEMAN":
-
-            v = self.valueTurnbullWakeman(valueDate,
-                                          stockPrice,
-                                          discountCurve,
-                                          dividendYield,
-                                          model,
-                                          accruedAverage)
-
-        elif valuationMethod == "CURRAN":
-
-            v = self.valueCurran(valueDate,
-                                 stockPrice,
-                                 discountCurve,
-                                 dividendYield,
-                                 model,
-                                 accruedAverage)
-
-#        elif valuationMethod == "MILEVSKY_POSNER":
-#
-#                v = self.valueMilevskyPosner(valueDate,
-#                                   stockPrice,
-#                                   dividendYield,
-#                                   volatility,
-#                                   interestRate,
-#                                   accruedAverage)
-#
-#        elif valuationMethod == "LEVY":
-#
-#                v = self.valueLevy(valueDate,
-#                                   stockPrice,
-#                                   dividendYield,
-#                                   volatility,
-#                                   interestRate,
-#                                   accruedAverage)
+        elif method == FinAsianOptionValuationMethods.TURNBULL_WAKEMAN:
+            v = self._valueTurnbullWakeman(valueDate,
+                                           stockPrice,
+                                           discountCurve,
+                                           dividendYield,
+                                           model,
+                                           accruedAverage)
 
+        elif method == FinAsianOptionValuationMethods.CURRAN:
+            v = self._valueCurran(valueDate,
+                                  stockPrice,
+                                  discountCurve,
+                                  dividendYield,
+                                  model,
+                                  accruedAverage)
         else:
             raise FinError("Unknown valuation model")
 
@@ -403,30 +421,23 @@ class FinEquityAsianOption(FinEquityOption):
 
 ###############################################################################
 
-    def valueGeometric(self,
-                       valueDate,
-                       stockPrice,
-                       discountCurve,
-                       dividendYield,
-                       model,
-                       accruedAverage):
-
-        # This is based on paper by Kemna and Vorst 1990. It calculates the
-        # Geometric Asian option price which is a lower bound on the Arithmetic
-        # option price.
-        # This should not be used as a valuation model for the Arithmetic Average
-        # option but can be used as a control variate for other approaches
+    def _valueGeometric(self, valueDate, stockPrice, discountCurve,
+                        dividendYield, model, accruedAverage):
+        ''' This option valuation is based on paper by Kemna and Vorst 1990. It
+        calculates the Geometric Asian option price which is a lower bound on
+        the Arithmetic option price. This should not be used as a valuation
+        model for the Arithmetic Average option but can be used as a control
+        variate for other approaches. '''
 
         if valueDate > self._expiryDate:
             raise FinError("Value date after option expiry date.")
 
         # the years to the start of the averaging period
         t0 = (self._startAveragingDate - valueDate) / gDaysInYear
-        t = (self._expiryDate - valueDate)  / gDaysInYear
+        t = (self._expiryDate - valueDate) / gDaysInYear
         tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
 
-        df = discountCurve.df(t)
-        r = -log(df)/t
+        r = discountCurve.zeroRate(self._expiryDate)
         volatility = model._volatility
 
         K = self._strikePrice
@@ -439,8 +450,7 @@ class FinEquityAsianOption(FinEquityOption):
         if t0 < 0:  # we are in the averaging period
 
             if accruedAverage is None:
-                raise FinError(
-                "In averaging period you need to enter the accrued average")
+                raise FinError(errorStr)
 
             # we adjust the strike to account for the accrued coupon
             K = (K * tau + accruedAverage * t0) / t
@@ -454,18 +464,18 @@ class FinEquityAsianOption(FinEquityOption):
         sigSq = volatility ** 2
         meanGeo = (r - q - sigSq / 2.0) * (t0 + (t - t0) / 2.0)
         varGeo = sigSq * (t0 + (t - t0) * (2 * n - 1) / (6 * n))
-        EG = S0 * exp(meanGeo + varGeo / 2.0)
+        EG = S0 * np.exp(meanGeo + varGeo / 2.0)
 
-        d1 = (meanGeo + log(S0 / K) + varGeo) / sqrt(varGeo)
-        d2 = d1 - sqrt(varGeo)
+        d1 = (meanGeo + np.log(S0 / K) + varGeo) / np.sqrt(varGeo)
+        d2 = d1 - np.sqrt(varGeo)
 
         # the Geometric price is the lower bound
-        call_g = exp(-r * t) * (EG * N(d1) - K * N(d2))
+        call_g = np.exp(-r * t) * (EG * N(d1) - K * N(d2))
 
         if self._optionType == FinOptionTypes.EUROPEAN_CALL:
             v = call_g
         elif self._optionType == FinOptionTypes.EUROPEAN_PUT:
-            put_g = call_g - (EG - K) * exp(-r * t)
+            put_g = call_g - (EG - K) * np.exp(-r * t)
             v = put_g
         else:
             raise FinError("Unknown option type " + str(self._optionType))
@@ -475,27 +485,21 @@ class FinEquityAsianOption(FinEquityOption):
 
 ###############################################################################
 
-    def valueCurran(self,
-                    valueDate,
-                    stockPrice,
-                    discountCurve,
-                    dividendYield,
-                    model,
-                    accruedAverage):
+    def _valueCurran(self, valueDate, stockPrice, discountCurve,
+                     dividendYield, model, accruedAverage):
+        ''' Valuation of an Asian option using the result by Vorst. '''
 
-        # This is based on paper by Vorst
         if valueDate > self._expiryDate:
             raise FinError("Value date after option expiry date.")
 
         # the years to the start of the averaging period
         t0 = (self._startAveragingDate - valueDate) / gDaysInYear
-        t = (self._expiryDate - valueDate)  / gDaysInYear
+        t = (self._expiryDate - valueDate) / gDaysInYear
         tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
 
         multiplier = 1.0
 
-        df = discountCurve.df(t)
-        r = -log(df)/t
+        r = discountCurve.zeroRate(self._expiryDate)
         volatility = model._volatility
 
         S0 = stockPrice
@@ -508,8 +512,7 @@ class FinEquityAsianOption(FinEquityOption):
         if t0 < 0:  # we are in the averaging period
 
             if accruedAverage is None:
-                raise FinError(
-                "In averaging period you need to enter the accrued average")
+                raise FinError(errorStr)
 
             # we adjust the strike to account for the accrued coupon
             K = (K * tau + accruedAverage * t0) / t
@@ -521,41 +524,35 @@ class FinEquityAsianOption(FinEquityOption):
             n = int(n * t / tau + 0.5) + 1
 
         h = (t - t0) / (n - 1)
-        u = (1.0 - exp(b * h * n)) / (1.0 - exp(b * h))
-        w = (1.0 - exp((2 * b + sigma2) * h * n)) / \
-            (1.0 - exp((2 * b + sigma2) * h))
+        u = (1.0 - np.exp(b * h * n)) / (1.0 - np.exp(b * h))
+        w = (1.0 - np.exp((2 * b + sigma2) * h * n)) / \
+            (1.0 - np.exp((2 * b + sigma2) * h))
 
-        FA = (S0 / n) * exp(b * t0) * u
-        EA2 = (S0 * S0 / n / n) * exp((2.0 * b + sigma2) * t0)
-        EA2 = EA2 * (w + 2.0 / (1 - exp((b + sigma2) * h)) * (u - w))
-        sigmaA = sqrt((log(EA2) - 2.0 * log(FA)) / t)
+        FA = (S0 / n) * np.exp(b * t0) * u
+        EA2 = (S0 * S0 / n / n) * np.exp((2.0 * b + sigma2) * t0)
+        EA2 = EA2 * (w + 2.0 / (1.0 - np.exp((b + sigma2) * h)) * (u - w))
+        sigmaA = np.sqrt((np.log(EA2) - 2.0 * np.log(FA)) / t)
 
-        d1 = (log(FA / K) + sigmaA * sigmaA * t / 2.0) / (sigmaA * sqrt(t))
-        d2 = d1 - sigmaA * sqrt(t)
+        d1 = (np.log(FA / K) + sigmaA * sigmaA * t / 2.0) / (sigmaA*np.sqrt(t))
+        d2 = d1 - sigmaA * np.sqrt(t)
 
         if self._optionType == FinOptionTypes.EUROPEAN_CALL:
-            v = exp(-r * t) * (FA * N(d1) - K * N(d2))
+            v = np.exp(-r * t) * (FA * N(d1) - K * N(d2))
         elif self._optionType == FinOptionTypes.EUROPEAN_PUT:
-            v = exp(-r * t) * (K * N(-d2) - FA * N(-d1))
+            v = np.exp(-r * t) * (K * N(-d2) - FA * N(-d1))
         else:
             return None
 
         v = v * multiplier
         return v
 
-##########################################################################
+###############################################################################
 
-    def valueTurnbullWakeman(self,
-                             valueDate,
-                             stockPrice,
-                             discountCurve,
-                             dividendYield,
-                             model,
-                             accruedAverage):
-
-        # This is based on paper by Turnbull and Wakeman 1991 which uses
-        # the edgeworth expansion to find the first two moments of the
-        # arithmetic average
+    def _valueTurnbullWakeman(self, valueDate, stockPrice, discountCurve,
+                              dividendYield, model, accruedAverage):
+        ''' Asian option valuation based on paper by Turnbull and Wakeman 1991
+        which uses the edgeworth expansion to find the first two moments of the
+        arithmetic average. '''
 
         if valueDate > self._expiryDate:
             raise FinError("Value date after option expiry date.")
@@ -568,16 +565,14 @@ class FinEquityAsianOption(FinEquityOption):
         multiplier = 1.0
         n = self._numObservations
 
-        df = discountCurve.df(t)
-        r = -log(df)/t
+        r = discountCurve.zeroRate(self._expiryDate)
 
         volatility = model._volatility
 
         if t0 < 0:  # we are in the averaging period
 
             if accruedAverage is None:
-                raise FinError(
-                "In averaging period you need to enter the accrued average")
+                raise FinError(errorStr)
 
             # we adjust the strike to account for the accrued coupon
             K = (K * tau + accruedAverage * t0) / t
@@ -599,155 +594,183 @@ class FinEquityAsianOption(FinEquityOption):
 
         if b == 0:
             M1 = 1.0
-            M2 = 2.0 * exp(sigma2 * t) - 2.0 * \
-                exp(sigma2 * t0) * (1.0 + sigma2 * dt)
+            M2 = 2.0 * np.exp(sigma2 * t) - 2.0 * \
+                np.exp(sigma2 * t0) * (1.0 + sigma2 * dt)
             M2 = M2 / sigma2 / sigma2 / dt / dt
         else:
-            M1 = S0 * (exp(b * t) - exp(b * t0)) / (b * dt)
-            M2 = exp(a2 * t) / a1 / a2 / dt / dt + \
-                (exp(a2 * t0) / b / dt / dt) * (1.0 / a2 - exp(b * dt) / a1)
+            M1 = S0 * (np.exp(b * t) - np.exp(b * t0)) / (b * dt)
+            M2 = np.exp(a2 * t) / a1 / a2 / dt / dt + \
+                (np.exp(a2 * t0) / b / dt / dt) * (1.0/a2 - np.exp(b*dt) / a1)
             M2 = 2.0 * M2 * S0 * S0
 
         F0 = M1
-        sigma2 = 1.0 / t * log(M2 / M1 / M1)
-        sigma = sqrt(sigma2)
+        sigma2 = 1.0 / t * np.log(M2 / M1 / M1)
+        sigma = np.sqrt(sigma2)
 
-        d1 = (log(F0 / K) + sigma2 * t / 2) / sigma / sqrt(t)
-        d2 = d1 - sigma * sqrt(t)
+        d1 = (np.log(F0 / K) + sigma2 * t / 2) / sigma / np.sqrt(t)
+        d2 = d1 - sigma * np.sqrt(t)
 
         if self._optionType == FinOptionTypes.EUROPEAN_CALL:
-            call = exp(-r * t) * (F0 * N(d1) - K * N(d2))
+            call = np.exp(-r * t) * (F0 * N(d1) - K * N(d2))
             v = call
         elif self._optionType == FinOptionTypes.EUROPEAN_PUT:
-            put = exp(-r * t) * (K * N(-d2) - F0 * N(-d1))
+            put = np.exp(-r * t) * (K * N(-d2) - F0 * N(-d1))
             v = put
         else:
             return None
 
         v = v * multiplier
-
         return v
 
-##############################################################################
+###############################################################################
 
-    def valueMC(self,
-                valueDate,
-                stockPrice,
-                discountCurve,
-                dividendYield,
-                model,
-                numPaths,
-                seed,
-                accruedAverage):
+    def _valueMC(self,
+                 valueDate: FinDate,
+                 stockPrice: float,
+                 discountCurve: FinDiscountCurve,
+                 dividendYield: float,
+                 model,
+                 numPaths: int,
+                 seed: int,
+                 accruedAverage: float):
+        ''' Monte Carlo valuation of the Asian Average option using standard
+        Monte Carlo code enhanced by Numba. I have discontinued the use of this
+        as it is both slow and has limited variance reduction. '''
 
         # Basic validation
         if valueDate > self._expiryDate:
             raise FinError("Value date after option expiry date.")
 
         if valueDate > self._startAveragingDate and accruedAverage is None:
-            raise FinError(
-                "In averaging period so need to enter accrued average.")
+            raise FinError(errorStr)
 
         # the years to the start of the averaging period
         t0 = (self._startAveragingDate - valueDate) / gDaysInYear
-        t = (self._expiryDate - valueDate)  / gDaysInYear
+        t = (self._expiryDate - valueDate) / gDaysInYear
         tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
 
-        df = discountCurve.df(t)
-        r = -log(df)/t
+        r = discountCurve.zeroRate(self._expiryDate)
+
         volatility = model._volatility
 
         K = self._strikePrice
         n = self._numObservations
 
-        v = valueMC_NUMBA(t0, t, tau, K, n, self._optionType,
-                          stockPrice,
-                          r,
-                          dividendYield,
-                          volatility,
-                          numPaths,
-                          seed,
-                          accruedAverage)
+        v = _valueMC_NUMBA(t0, t, tau, K, n, self._optionType,
+                           stockPrice,
+                           r,
+                           dividendYield,
+                           volatility,
+                           numPaths,
+                           seed,
+                           accruedAverage)
+
+        return v
+
+##############################################################################
+
+    def _valueMC_fast(self,
+                      valueDate,
+                      stockPrice,
+                      discountCurve,
+                      dividendYield,  # Yield
+                      model,          # Model
+                      numPaths,       # Numpaths integer
+                      seed,
+                      accruedAverage):
+        ''' Monte Carlo valuation of the Asian Average option. This method uses
+        a lot of Numpy vectorisation. It is also helped by Numba. '''
+
+        # the years to the start of the averaging period
+        t0 = (self._startAveragingDate - valueDate) / gDaysInYear
+        t = (self._expiryDate - valueDate) / gDaysInYear
+        tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
+
+        K = self._strikePrice
+        n = self._numObservations
+
+        r = discountCurve.zeroRate(self._expiryDate)
+
+        volatility = model._volatility
+
+        v = _valueMC_fast_NUMBA(t0, t, tau,
+                                K, n, self._optionType,
+                                stockPrice,
+                                r,
+                                dividendYield,
+                                volatility,
+                                numPaths,
+                                seed,
+                                accruedAverage)
 
         return v
 
 ###############################################################################
 
-    def valueMC_fast(self,
-                     valueDate,
-                     stockPrice,
-                     discountCurve,
-                     dividendYield,
-                     model,
-                     numPaths,
-                     seed,
-                     accruedAverage):
+    def valueMC(self,
+                valueDate: FinDate,
+                stockPrice: float,
+                discountCurve: FinDiscountCurve,
+                dividendYield: float,
+                model,
+                numPaths: int,
+                seed: int,
+                accruedAverage: float):
+        ''' Monte Carlo valuation of the Asian Average option using a control
+        variate method that improves accuracy and reduces the variance of the
+        price. This uses Numpy and Numba. This is the standard MC pricer. '''
 
         # the years to the start of the averaging period
         t0 = (self._startAveragingDate - valueDate) / gDaysInYear
-        t = (self._expiryDate - valueDate)  / gDaysInYear
+        t = (self._expiryDate - valueDate) / gDaysInYear
         tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
 
         K = self._strikePrice
         n = self._numObservations
 
-        df = discountCurve.df(t)
-        r = -log(df)/t
-        volatility = model._volatility
-
-        v = valueMC_fast_NUMBA(t0, t, tau, K, n, self._optionType,
-                               stockPrice,
-                               r,
-                               dividendYield,
-                               volatility,
-                               numPaths,
-                               seed,
-                               accruedAverage)
-
-        return v
-
-##########################################################################
-
-    def valueMC_fast_CV(self,
-                        valueDate,
-                        stockPrice,
-                        discountCurve,
-                        dividendYield,
-                        model,
-                        numPaths,
-                        seed,
-                        accruedAverage):
-
-        # the years to the start of the averaging period
-        t0 = (self._startAveragingDate - valueDate) / gDaysInYear
-        t = (self._expiryDate - valueDate)  / gDaysInYear
-        tau = (self._expiryDate - self._startAveragingDate) / gDaysInYear
-
-        K = self._strikePrice
-        n = self._numObservations
-
-        df = discountCurve.df(t)
-        r = -log(df)/t
+        r = discountCurve.zeroRate(self._expiryDate)
         volatility = model._volatility
 
         # For control variate we price a Geometric average option exactly
-        v_g_exact = self.valueGeometric(valueDate,
-                                        stockPrice,
-                                        discountCurve,
-                                        dividendYield,
-                                        model,
-                                        accruedAverage)
+        v_g_exact = self._valueGeometric(valueDate,
+                                         stockPrice,
+                                         discountCurve,
+                                         dividendYield,
+                                         model,
+                                         accruedAverage)
 
-        v = valueMC_fast_CV_NUMBA(t0, t, tau, K, n, self._optionType,
-                                  stockPrice,
-                                  r,
-                                  dividendYield,
-                                  volatility,
-                                  numPaths,
-                                  seed,
-                                  accruedAverage,
-                                  v_g_exact)
+        v = _valueMC_fast_CV_NUMBA(t0,
+                                   t,
+                                   tau,
+                                   K,
+                                   n,
+                                   self._optionType,
+                                   stockPrice,
+                                   r,
+                                   dividendYield,
+                                   volatility,
+                                   numPaths,
+                                   seed,
+                                   accruedAverage,
+                                   v_g_exact)
 
         return v
 
-##########################################################################
+###############################################################################
+
+    def __repr__(self):
+        s = labelToString("OBJECT TYPE", type(self).__name__)
+        s += labelToString("START AVERAGING DATE", self._startAveragingDate)
+        s += labelToString("EXPIRY DATE", self._expiryDate)
+        s += labelToString("STRIKE PRICE", self._strikePrice)
+        s += labelToString("OPTION TYPE", self._optionType)
+        s += labelToString("NUM OBSERVATIONS", self._numObservations, "")
+        return s
+
+###############################################################################
+
+    def _print(self):
+        ''' Simple print function for backward compatibility. '''
+        print(self)
+
+###############################################################################
