@@ -1,24 +1,23 @@
-# -*- coding: utf-8 -*-
-"""
-Created on Fri Feb 12 16:51:05 2016
-
-@author: Dominic O'Kane
-"""
+##############################################################################
+# Copyright (C) 2018, 2019, 2020 Dominic O'Kane
+##############################################################################
 
 import numpy as np
 from scipy import optimize
 from scipy.stats import norm
 
+
 from ...finutils.FinDate import FinDate
 from ...finutils.FinMath import nprime
 from ...finutils.FinGlobalVariables import gDaysInYear
 from ...finutils.FinError import FinError
-from ...products.FinOptionTypes import FinOptionTypes
+from ...finutils.FinOptionTypes import FinOptionTypes
 from ...products.fx.FinFXModelTypes import FinFXModel
 from ...products.fx.FinFXModelTypes import FinFXModelBlackScholes
 from ...products.fx.FinFXModelTypes import FinFXModelSABR
 from ...models.FinModelCRRTree import crrTreeValAvg
 from ...models.FinModelSABR import blackVolFromSABR
+from ...finutils.FinHelperFunctions import checkArgumentTypes
 
 N = norm.cdf
 
@@ -122,7 +121,7 @@ def solveForStrike(valueDate,
                 forDiscountCurve, delta, deltaType, volatility)
 
     sigma = optimize.newton(g, x0=spotFXRate, args=argtuple,
-                            tol=1e-5, maxiter=50, fprime2=None)
+                            tol=1e-5, maxiter=100, fprime2=None)
     return sigma
 
 ###############################################################################
@@ -140,13 +139,13 @@ class FinFXVanillaOption():
     well as the various Greek risk sensitivies. '''
 
     def __init__(self,
-                 expiryDate,
-                 strikeFXRate,  # ONE UNIT OF FOREIGN IN DOMESTIC CCY
-                 currencyPair,  # FORDOM
-                 optionType,
-                 notional,
-                 premCurrency,
-                 spotDays=0):
+                 expiryDate: FinDate,
+                 strikeFXRate: float,  # 1 unit of foreign in domestic
+                 currencyPair: str,  # FORDOM
+                 optionType: FinOptionTypes,
+                 notional: float,
+                 premCurrency: str,
+                 spotDays: int = 0):
         ''' Create the FX Vanilla Option object. Inputs include expiry date,
         strike, currency pair, option type (call or put), notional and the
         currency of the notional. And adjustment for spot days is enabled. All
@@ -155,6 +154,8 @@ class FinFXVanillaOption():
         where FOR is the foreign currency pair currency code and DOM is the
         same for the domestic currency. '''
 
+        checkArgumentTypes(self.__init__, locals())
+
         deliveryDate = expiryDate.addWorkDays(spotDays)
 
         ''' The FX rate the price in domestic currency ccy2 of a single unit
@@ -199,7 +200,7 @@ class FinFXVanillaOption():
 
     def value(self,
               valueDate,
-              spotFXRate,  # ONE UNIT OF FOREIGN IN DOMESTIC CCY
+              spotFXRate,  # 1 unit of foreign in domestic
               domDiscountCurve,
               forDiscountCurve,
               model):
@@ -227,18 +228,15 @@ class FinFXVanillaOption():
 
         tdel = np.maximum(tdel, 1e-10)
 
-        domDF = domDiscountCurve.df(tdel)
-        rd = -np.log(domDF)/tdel
-
-        forDF = forDiscountCurve.df(tdel)
-        rf = -np.log(forDF)/tdel
+        rd = domDiscountCurve.zeroRate(self._deliveryDate)
+        rf = forDiscountCurve.zeroRate(self._deliveryDate)
 
         S0 = spotFXRate
         K = self._strikeFXRate
         F0T = S0 * np.exp((rd-rf)*tdel)
 
         if type(model) == FinFXModelBlackScholes or \
-            type(model) == FinFXModelSABR:
+           type(model) == FinFXModelSABR:
 
             if type(model) == FinFXModelBlackScholes:
                 volatility = model._volatility
@@ -270,11 +268,11 @@ class FinFXVanillaOption():
             elif self._optionType == FinOptionTypes.AMERICAN_CALL:
                 numStepsPerYear = 100
                 vdf = crrTreeValAvg(S0, rd, rf, volatility, numStepsPerYear,
-                        texp, FinOptionTypes.AMERICAN_CALL.value, K)['value']
+                          texp, FinOptionTypes.AMERICAN_CALL.value, K)['value']
             elif self._optionType == FinOptionTypes.AMERICAN_PUT:
                 numStepsPerYear = 100
                 vdf = crrTreeValAvg(S0, rd, rf, volatility, numStepsPerYear,
-                        texp, FinOptionTypes.AMERICAN_PUT.value, K)['value']
+                           texp, FinOptionTypes.AMERICAN_PUT.value, K)['value']
             else:
                 raise FinError("Unknown option type")
 
@@ -380,10 +378,10 @@ class FinFXVanillaOption():
 
         tdel = np.maximum(tdel, 1e-10)
 
-        domDf = domDiscountCurve.df(tdel)
+        domDf = domDiscountCurve._df(tdel)
         rd = -np.log(domDf)/tdel
 
-        forDf = forDiscountCurve.df(tdel)
+        forDf = forDiscountCurve._df(tdel)
         rf = -np.log(forDf)/tdel
 
         S0 = spotFXRate
@@ -415,10 +413,13 @@ class FinFXVanillaOption():
             else:
                 raise FinError("Unknown option type")
 
-            pips_spot_delta = w*np.exp(-rf*tdel)*N(w*d1)
-            pips_fwd_delta = w*N(w*d1)
-            pct_spot_delta_prem_adj = w*np.exp(-rd*tdel)*N(w*d2)*K/S0
-            pct_fwd_delta_prem_adj = w*K*N(w*d2)/F
+            n1 = N(w*d1)
+            n2 = N(w*d2)
+
+            pips_spot_delta = w*np.exp(-rf*tdel)*n1
+            pips_fwd_delta = w*n1
+            pct_spot_delta_prem_adj = w*np.exp(-rd*tdel)*n2*K/S0
+            pct_fwd_delta_prem_adj = w*K*n2/F
 
         return {"pips_spot_delta": pips_spot_delta,
                 "pips_fwd_delta": pips_fwd_delta,
@@ -452,10 +453,10 @@ class FinFXVanillaOption():
 
         t = np.maximum(t, 1e-10)
 
-        domDf = domDiscountCurve.df(t)
+        domDf = domDiscountCurve._df(t)
         rd = -np.log(domDf)/t
 
-        forDf = forDiscountCurve.df(t)
+        forDf = forDiscountCurve._df(t)
         rf = -np.log(forDf)/t
 
         K = self._strikeFXRate
@@ -510,10 +511,10 @@ class FinFXVanillaOption():
 
         t = np.maximum(t, 1e-10)
 
-        domDf = domDiscountCurve.df(t)
+        domDf = domDiscountCurve._df(t)
         rd = -np.log(domDf)/t
 
-        forDf = forDiscountCurve.df(t)
+        forDf = forDiscountCurve._df(t)
         rf = -np.log(forDf)/t
 
         K = self._strikeFXRate
@@ -566,10 +567,10 @@ class FinFXVanillaOption():
 
         t = np.maximum(t, 1e-10)
 
-        domDf = domDiscountCurve.df(t)
+        domDf = domDiscountCurve._df(t)
         rd = -np.log(domDf)/t
 
-        forDf = forDiscountCurve.df(t)
+        forDf = forDiscountCurve._df(t)
         rf = -np.log(forDf)/t
 
         K = self._strikeFXRate
